# Austin Griffith
# Barriers

import numpy as np
from scipy.stats import norm

# lambda constant
def _lambdaVal(r,vol,q):
    l = 1 + (r - q - vol*vol*0.5) / (vol*vol)
    return(l)

# I's are set values used as the analytical solution to the barrier options
def _I1(alpha,beta,s,k,r,Z,T,vol,q):
    L = _lambdaVal(r,vol,q)
    xval = np.log(s/k) / (vol*np.sqrt(T)) + L*vol*np.sqrt(T)
    partial = alpha*s*norm.cdf(alpha*xval) - alpha*k*np.exp(-r*T)*norm.cdf(alpha*xval - alpha*vol*np.sqrt(T))
    return(partial)

def _I2(alpha,beta,s,k,r,Z,T,vol,q):
    L = _lambdaVal(r,vol,q)
    xval = np.log(s/Z)/(vol*np.sqrt(T)) + L*vol*np.sqrt(T)
    partial = alpha*s*norm.cdf(alpha*xval) - alpha*k*np.exp(-r*T)*norm.cdf(alpha*xval - alpha*vol*np.sqrt(T))
    return(partial)

def _I3(alpha,beta,s,k,r,Z,T,vol,q):
    L = _lambdaVal(r,vol,q)
    xval = np.log(np.square(Z)/(s*k))/(vol*np.sqrt(T)) + L*vol*np.sqrt(T)
    partial = (alpha*s*np.power(Z/s,2*L)*norm.cdf(beta*xval) -
        alpha*k*np.exp(-r*T)*np.power(Z/s,2*L-2)*norm.cdf(beta*xval-beta*vol*np.sqrt(T)))
    return(partial)

def _I4(alpha,beta,s,k,r,Z,T,vol,q):
    L = _lambdaVal(r,vol,q)
    xval = np.log(Z/s)/(vol*np.sqrt(T)) + L*vol*np.sqrt(T)
    partial = (alpha*s*np.power(Z/s,2*L)*norm.cdf(beta*xval) -
        alpha*k*np.exp(-r*T)*np.power(Z/s,2*L - 2)*norm.cdf(beta*xval - beta*vol*np.sqrt(T)))
    return(partial)

# out barriers
def DownOutPutLow(s,k,r,Z,T,vol,q):
    '''
    Calculate the Down-and-Out Put option, where barrier is less than strike

    Parameters
    ----------
    s : number of any type (int, float8, float64 etc.)
        Spot value of underlying asset at current time, t
    k : number of any type (int, float8, float64 etc.)
        Strike value of option, determined at initiation
    r : number of any type (int, float8, float64 etc.)
        Risk free interest rate, implied constant till expiration
    Z : number of any type (int, float8, float64 etc.)
        Barrier value of option, determined at initiation
    T : number of any type (int, float8, float64 etc.)
        Time till expiration for option, can be interpreted as 'T - t' should
        the option already be initiated, and be 't' time from time = 0
    vol : number of any type (int, float8, float64 etc.)
        Volatility of underlying, implied constant till expiration in Black
        Scholes model
    q : number of any type (int, float8, float64 etc.)
        Continuous dividend payout, as a percentage

    * Z < k must hold true for this function

    Returns
    -------
    price : float
        Price value of barrier option.

    Examples
    --------
    >>> from qcfoptions.barriers import DownOutPutLow
    >>> s = 1
        k = 1
        r = 0.015
        Z = 0.6
        T = 2
        vol = 0.25
        q = 0.01
    >>> DownOutPutLow(s,k,r,Z,T,vol,q)
    0.05510635577234088

    '''
    a = -1
    b = 1

    if s >= Z:
        price = (_I1(a,b,s,k,r,Z,T,vol,q) - _I2(a,b,s,k,r,Z,T,vol,q) +
                _I3(a,b,s,k,r,Z,T,vol,q) - _I4(a,b,s,k,r,Z,T,vol,q))
    else:
        price = 0.0
    return(max(price, 0.0))

def DownOutCall(s,k,r,Z,T,vol,q):
    '''
    Calculate the Down-and-Out Call option, for any barrier

    Parameters
    ----------
    s : number of any type (int, float8, float64 etc.)
        Spot value of underlying asset at current time, t
    k : number of any type (int, float8, float64 etc.)
        Strike value of option, determined at initiation
    r : number of any type (int, float8, float64 etc.)
        Risk free interest rate, implied constant till expiration
    Z : number of any type (int, float8, float64 etc.)
        Barrier value of option, determined at initiation
    T : number of any type (int, float8, float64 etc.)
        Time till expiration for option, can be interpreted as 'T - t' should
        the option already be initiated, and be 't' time from time = 0
    vol : number of any type (int, float8, float64 etc.)
        Volatility of underlying, implied constant till expiration in Black
        Scholes model
    q : number of any type (int, float8, float64 etc.)
        Continuous dividend payout, as a percentage

    Returns
    -------
    price : float
        Price value of barrier option.

    Examples
    --------
    >>> from qcfoptions.barriers import DownOutCall
    >>> s = 1
        k = 1
        r = 0.015
        Z = 0.6
        T = 2
        vol = 0.25
        q = 0.01
    >>> DownOutCall(s,k,r,Z,T,vol,q)
    0.05510635577234088

    '''
    a = 1
    b = 1

    if k > Z and s >= Z:
        price = _I1(a,b,s,k,r,Z,T,vol,q) - _I3(a,b,s,k,r,Z,T,vol,q)
    elif k <= Z and s >= Z:
        price = _I2(a,b,s,k,r,Z,T,vol,q) - _I4(a,b,s,k,r,Z,T,vol,q)
    else:
        price = 0.0
    return(max(price, 0.0))

def UpOutCallHigh(s,k,r,Z,T,vol,q):
    '''
    Calculate the Up-and-Out Call option, where barrier is greater than strike

    Parameters
    ----------
    s : number of any type (int, float8, float64 etc.)
        Spot value of underlying asset at current time, t
    k : number of any type (int, float8, float64 etc.)
        Strike value of option, determined at initiation
    r : number of any type (int, float8, float64 etc.)
        Risk free interest rate, implied constant till expiration
    Z : number of any type (int, float8, float64 etc.)
        Barrier value of option, determined at initiation
    T : number of any type (int, float8, float64 etc.)
        Time till expiration for option, can be interpreted as 'T - t' should
        the option already be initiated, and be 't' time from time = 0
    vol : number of any type (int, float8, float64 etc.)
        Volatility of underlying, implied constant till expiration in Black
        Scholes model
    q : number of any type (int, float8, float64 etc.)
        Continuous dividend payout, as a percentage

    * Z > k must hold true for this function

    Returns
    -------
    price : float
        Price value of barrier option.

    Examples
    --------
    >>> from qcfoptions.barriers import UpOutCallHigh
    >>> s = 1
        k = 1
        r = 0.015
        Z = 1.2
        T = 2
        vol = 0.25
        q = 0.01
    >>> UpOutCallHigh(s,k,r,Z,T,vol,q)
    0.003623404694383797

    '''
    a = 1
    b = -1

    if s <= Z:
        price = (_I1(a,b,s,k,r,Z,T,vol,q) - _I2(a,b,s,k,r,Z,T,vol,q) +
                _I3(a,b,s,k,r,Z,T,vol,q) - _I4(a,b,s,k,r,Z,T,vol,q))
    else:
        price = 0.0
    return(max(price, 0.0))

def UpOutPut(s,k,r,Z,T,vol,q):
    '''
    Calculate the Up-and-Out Put option, for any barrier

    Parameters
    ----------
    s : number of any type (int, float8, float64 etc.)
        Spot value of underlying asset at current time, t
    k : number of any type (int, float8, float64 etc.)
        Strike value of option, determined at initiation
    r : number of any type (int, float8, float64 etc.)
        Risk free interest rate, implied constant till expiration
    Z : number of any type (int, float8, float64 etc.)
        Barrier value of option, determined at initiation
    T : number of any type (int, float8, float64 etc.)
        Time till expiration for option, can be interpreted as 'T - t' should
        the option already be initiated, and be 't' time from time = 0
    vol : number of any type (int, float8, float64 etc.)
        Volatility of underlying, implied constant till expiration in Black
        Scholes model
    q : number of any type (int, float8, float64 etc.)
        Continuous dividend payout, as a percentage

    Returns
    -------
    price : float
        Price value of barrier option.

    Examples
    --------
    >>> from qcfoptions.barriers import UpOutPut
    >>> s = 1
        k = 1
        r = 0.015
        Z = 1.2
        T = 2
        vol = 0.25
        q = 0.01
    >>> UpOutPut(s,k,r,Z,T,vol,q)
    0.10028057349038572

    '''
    a = -1
    b = -1

    if k <= Z and s <= Z:
        price = _I1(a,b,s,k,r,Z,T,vol,q) - _I3(a,b,s,k,r,Z,T,vol,q)
    elif k > Z and s <= Z:
        price = _I2(a,b,s,k,r,Z,T,vol,q) - _I4(a,b,s,k,r,Z,T,vol,q)
    else:
        price = 0.0
    return(max(price, 0.0))

# in barriers
def DownInCall(s,k,r,Z,T,vol,q):
    '''
    Calculate the Down-and-In Call option, for any barrier

    Parameters
    ----------
    s : number of any type (int, float8, float64 etc.)
        Spot value of underlying asset at current time, t
    k : number of any type (int, float8, float64 etc.)
        Strike value of option, determined at initiation
    r : number of any type (int, float8, float64 etc.)
        Risk free interest rate, implied constant till expiration
    Z : number of any type (int, float8, float64 etc.)
        Barrier value of option, determined at initiation
    T : number of any type (int, float8, float64 etc.)
        Time till expiration for option, can be interpreted as 'T - t' should
        the option already be initiated, and be 't' time from time = 0
    vol : number of any type (int, float8, float64 etc.)
        Volatility of underlying, implied constant till expiration in Black
        Scholes model
    q : number of any type (int, float8, float64 etc.)
        Continuous dividend payout, as a percentage

    Returns
    -------
    price : float
        Price value of barrier option.

    Examples
    --------
    >>> from qcfoptions.barriers import DownInCall
    >>> s = 1
        k = 1
        r = 0.015
        Z = 1.2
        T = 2
        vol = 0.25
        q = 0.01
    >>> DownInCall(s,k,r,Z,T,vol,q)
    0.4299600426787612

    '''
    a = 1
    b = 1

    if k > Z:
        price = _I3(a,b,s,k,r,Z,T,vol,q)
    elif k <= Z:
        price = (_I1(a,b,s,k,r,Z,T,vol,q) - _I2(a,b,s,k,r,Z,T,vol,q) +
            _I4(a,b,s,k,r,Z,T,vol,q))
    else:
        price = 0.0
    return(max(price, 0.0))

def DownInPut(s,k,r,Z,T,vol,q):
    '''
    Calculate the Down-and-In Put option, for any barrier

    Parameters
    ----------
    s : number of any type (int, float8, float64 etc.)
        Spot value of underlying asset at current time, t
    k : number of any type (int, float8, float64 etc.)
        Strike value of option, determined at initiation
    r : number of any type (int, float8, float64 etc.)
        Risk free interest rate, implied constant till expiration
    Z : number of any type (int, float8, float64 etc.)
        Barrier value of option, determined at initiation
    T : number of any type (int, float8, float64 etc.)
        Time till expiration for option, can be interpreted as 'T - t' should
        the option already be initiated, and be 't' time from time = 0
    vol : number of any type (int, float8, float64 etc.)
        Volatility of underlying, implied constant till expiration in Black
        Scholes model
    q : number of any type (int, float8, float64 etc.)
        Continuous dividend payout, as a percentage

    Returns
    -------
    price : float
        Price value of barrier option.

    Examples
    --------
    >>> from qcfoptions.barriers import DownInPut
    >>> s = 1
        k = 1
        r = 0.015
        Z = 1.2
        T = 2
        vol = 0.25
        q = 0.01
    >>> DownInPut(s,k,r,Z,T,vol,q)
    0.12373904745993392

    '''
    a = -1
    b = 1

    if k > Z:
        price = (_I2(a,b,s,k,r,Z,T,vol,q) - _I3(a,b,s,k,r,Z,T,vol,q) +
            _I4(a,b,s,k,r,Z,T,vol,q))
    elif k <= Z:
        price = _I1(a,b,s,k,r,Z,T,vol,q)
    else:
        price = 0.0
    return(max(price, 0.0))

def UpInCall(s,k,r,Z,T,vol,q):
    '''
    Calculate the Up-and-In Call option, for any barrier

    Parameters
    ----------
    s : number of any type (int, float8, float64 etc.)
        Spot value of underlying asset at current time, t
    k : number of any type (int, float8, float64 etc.)
        Strike value of option, determined at initiation
    r : number of any type (int, float8, float64 etc.)
        Risk free interest rate, implied constant till expiration
    Z : number of any type (int, float8, float64 etc.)
        Barrier value of option, determined at initiation
    T : number of any type (int, float8, float64 etc.)
        Time till expiration for option, can be interpreted as 'T - t' should
        the option already be initiated, and be 't' time from time = 0
    vol : number of any type (int, float8, float64 etc.)
        Volatility of underlying, implied constant till expiration in Black
        Scholes model
    q : number of any type (int, float8, float64 etc.)
        Continuous dividend payout, as a percentage

    Returns
    -------
    price : float
        Price value of barrier option.

    Examples
    --------
    >>> from qcfoptions.barriers import UpInCall
    >>> s = 1
        k = 1
        r = 0.015
        Z = 1.2
        T = 2
        vol = 0.25
        q = 0.01
    >>> UpInCall(s,k,r,Z,T,vol,q)
    0.14967010921704207

    '''
    a = 1
    b = -1

    if k > Z:
        price = _I1(a,b,s,k,r,Z,T,vol,q)
    elif k <= Z:
        price = (_I2(a,b,s,k,r,Z,T,vol,q) - _I3(a,b,s,k,r,Z,T,vol,q) +
            _I4(a,b,s,k,r,Z,T,vol,q))
    else:
        price = 0.0
    return(max(price, 0.0))

def UpInPut(s,k,r,Z,T,vol,q):
    '''
    Calculate the Up-and-In Put option, for any barrier

    Parameters
    ----------
    s : number of any type (int, float8, float64 etc.)
        Spot value of underlying asset at current time, t
    k : number of any type (int, float8, float64 etc.)
        Strike value of option, determined at initiation
    r : number of any type (int, float8, float64 etc.)
        Risk free interest rate, implied constant till expiration
    Z : number of any type (int, float8, float64 etc.)
        Barrier value of option, determined at initiation
    T : number of any type (int, float8, float64 etc.)
        Time till expiration for option, can be interpreted as 'T - t' should
        the option already be initiated, and be 't' time from time = 0
    vol : number of any type (int, float8, float64 etc.)
        Volatility of underlying, implied constant till expiration in Black
        Scholes model
    q : number of any type (int, float8, float64 etc.)
        Continuous dividend payout, as a percentage

    Returns
    -------
    price : float
        Price value of barrier option.

    Examples
    --------
    >>> from qcfoptions.barriers import UpInCall
    >>> s = 1
        k = 1
        r = 0.015
        Z = 1.2
        T = 2
        vol = 0.25
        q = 0.01
    >>> UpInPut(s,k,r,Z,T,vol,q)
    0.0234584739695482

    '''
    a = -1
    b = -1

    if k > Z:
        price = (_I1(a,b,s,k,r,Z,T,vol,q) - _I2(a,b,s,k,r,Z,T,vol,q) +
            _I4(a,b,s,k,r,Z,T,vol,q))
    elif k <= Z:
        price = _I3(a,b,s,k,r,Z,T,vol,q)
    else:
        price = 0.0
    return(max(price, 0.0))



# Austin Griffith
# Options, Black Scholes

import numpy as np
from scipy.stats import norm

def Euro(s,k,r,T,vol,q):
    '''
    Calculate the Black Scholes value of a European Call / Put option
    Payoffs are of the form :
    C = max(S - K, 0)
    P = max(K - S, 0)

    Parameters
    ----------
    s : number of any type (int, float8, float64 etc.), numpy array of any type
        should the user wish to have a list of values output with varying s
        Spot value of underlying asset at current time, t
    k : number of any type (int, float8, float64 etc.), numpy array of any type
        should the user wish to have a list of values output with varying k
        Strike value of option, determined at initiation
    r : number of any type (int, float8, float64 etc.), numpy array of any type
        should the user wish to have a list of values output with varying r
        Risk free interest rate, implied constant till expiration
    T : number of any type (int, float8, float64 etc.), numpy array of any type
        should the user wish to have a list of values output with varying T
        Time till expiration for option
    vol : number of any type (int, float8, float64 etc.), numpy array of any type
        should the user wish to have a list of values output with varying vol
        Volatility of underlying, implied constant till expiration in Black
        Scholes model
    q : number of any type (int, float8, float64 etc.), numpy array of any type
        should the user wish to have a list of values output with varying q
        Continuous dividend payout, as a percentage

    All parameters can be individual values.
    Only one of these parameters can be a numpy.array, otherwise there will be
    a dimension mismatch.

    Returns
    -------
    [call,put] : list of pair of float or numpy.array values
        European call and put values, type depends on input value.
        If all input values are individual numbers, then output will be float.
        If one input value is numpy.array, then output will be numpy.array.

    Examples
    --------
    >>> from qcfoptions.bsoptions import Euro
    >>> s = 1
        k = 1
        r = 0.015
        T = 2
        vol = 0.25
        q = 0.01
    >>> Euro(s,k,r,T,vol,q)
    [0.14178423526329242, 0.13203109550504533]
    >>> import numpy as np
    >>> s_array = np.array([0.5,1.0,1.5,2.0])
    >>> Euro(s_array,k,r,T,vol,q)
    [array([0.00244785, 0.14178424, 0.52452167, 0.99416626]),
    array([0.48279404, 0.1320311 , 0.0246692 , 0.00421444])]

    '''
    d1 = ((np.log(s/k) + (r - q + 0.5*vol*vol)*T)) / (vol*np.sqrt(T))
    d2 = d1 - vol*np.sqrt(T)

    option = np.exp(-q*T)*s
    strike = np.exp(-r*T)*k

    put = strike*norm.cdf(-d2) - option*norm.cdf(-d1)
    call = option*norm.cdf(d1) - strike*norm.cdf(d2)
    return([call,put])

def AsianGeometric(s,k,r,T,vol,q):
    '''
    Calculate the Black Scholes value of Geometric Average Asian Call / Put
    option with a fixed strike
    Payoffs are of the form :
    C = max(AVG_geo - K, 0)
    P = max(K - AVG_geo, 0)

    Parameters
    ----------
    s : number of any type (int, float8, float64 etc.), numpy array of any type
        should the user wish to have a list of values output with varying s
        Spot value of underlying asset at current time, t
    k : number of any type (int, float8, float64 etc.), numpy array of any type
        should the user wish to have a list of values output with varying k
        Strike value of option, determined at initiation
    r : number of any type (int, float8, float64 etc.), numpy array of any type
        should the user wish to have a list of values output with varying r
        Risk free interest rate, implied constant till expiration
    T : number of any type (int, float8, float64 etc.), numpy array of any type
        should the user wish to have a list of values output with varying T
        Time till expiration for option
    vol : number of any type (int, float8, float64 etc.), numpy array of any type
        should the user wish to have a list of values output with varying vol
        Volatility of underlying, implied constant till expiration in Black
        Scholes model
    q : number of any type (int, float8, float64 etc.), numpy array of any type
        should the user wish to have a list of values output with varying q
        Continuous dividend payout, as a percentage

    All parameters can be individual values.
    Only one of these parameters can be a numpy.array, otherwise there will be
    a dimension mismatch.

    Returns
    -------
    [call,put] : list of pair of float or numpy.array values
        Asian call and put values, type depends on input value.
        If all input values are individual numbers, then output will be float.
        If one input value is numpy.array, then output will be numpy.array.

    Examples
    --------
    >>> from qcfoptions.bsoptions import AsianGeometric
    >>> s = 1
        k = 1
        r = 0.015
        T = 2
        vol = 0.25
        q = 0.01
    >>> AsianGeometric(s,k,r,T,vol,q)
    [0.0760833709740742, 0.08132574004684912]
    >>> import numpy as np
    >>> s_array = np.array([0.5,1.0,1.5,2.0])
    >>> AsianGeometric(s_array,k,r,T,vol,q)
    [array([1.10488795e-05, 7.60833710e-02, 4.79636239e-01, 9.59987952e-01]),
    array([4.87855000e-01, 8.13257400e-02, 2.27702612e-03, 2.71568147e-05])]

    '''
    a = 0.5*(r - q - vol*vol/6)
    volG = vol/np.sqrt(3)

    d1 = (np.log(s/k) + (a + 0.5*volG*volG)*T) / (volG*np.sqrt(T))
    d2 = d1 - volG*np.sqrt(T)

    option = s*np.exp((a - r)*T)
    strike = k*np.exp(-r*T)

    put = strike*norm.cdf(-d2) - option*norm.cdf(-d1)
    call = option*norm.cdf(d1) - strike*norm.cdf(d2)
    return([call,put])

def AsianArithmetic(s,k,r,T,vol,q):
    '''
    Calculate the Black Scholes value of Arithmetic Average Asian Call / Put
    option with a fixed strike
    Payoffs are of the form :
    C = max(AVG_arithmetic - K, 0)
    P = max(K - AVG_arithmetic, 0)

    Parameters
    ----------
    s : number of any type (int, float8, float64 etc.), numpy array of any type
        should the user wish to have a list of values output with varying s
        Spot value of underlying asset at current time, t
    k : number of any type (int, float8, float64 etc.), numpy array of any type
        should the user wish to have a list of values output with varying k
        Strike value of option, determined at initiation
    r : number of any type (int, float8, float64 etc.), numpy array of any type
        should the user wish to have a list of values output with varying r
        Risk free interest rate, implied constant till expiration
    T : number of any type (int, float8, float64 etc.), numpy array of any type
        should the user wish to have a list of values output with varying T
        Time till expiration for option
    vol : number of any type (int, float8, float64 etc.), numpy array of any type
        should the user wish to have a list of values output with varying vol
        Volatility of underlying, implied constant till expiration in Black
        Scholes model
    q : number of any type (int, float8, float64 etc.), numpy array of any type
        should the user wish to have a list of values output with varying q
        Continuous dividend payout, as a percentage

    All parameters can be individual values.
    Only one of these parameters can be a numpy.array, otherwise there will be
    a dimension mismatch.

    * r > q, else the natural logarithm will break due to a relationship
        mismatch

    Returns
    -------
    [call,put] : list of pair of float or numpy.array values
        Asian call and put values, type depends on input value.
        If all input values are individual numbers, then output will be float.
        If one input value is numpy.array, then output will be numpy.array.

    Examples
    --------
    >>> from qcfoptions.bsoptions import AsianArithmetic
    >>> s = 1
        k = 1
        r = 0.01
        T = 2
        vol = 0.25
        q = 0.015
    >>> AsianArithmetic(s,k,r,T,vol,q)
    [0.8242269391820242, 0.8193584969058301]
    >>> import numpy as np
    >>> s_array = np.array([0.5,1.0,1.5,2.0])
    >>> AsianArithmetic(s_array,k,r,T,vol,q)
    [array([0.38280732, 0.82422694, 1.27915927, 1.7410437 ]),
    array([0.86559587, 0.8193585 , 0.78663384, 0.76086128])]

    '''
    m1 = s*(np.exp((r - q)*T) - 1) / ((r - q)*T)
    m2l = 2*s*s*np.exp((2*r - 2*q + vol*vol)*T) / ((r - q +vol*vol)*T*T*(2*r - 2*q + vol*vol))
    m2r = (2*s*s / ((r-q)*T*T))*((1/(2*(r-q) + vol*vol)) -
        np.exp((r-q)*T)/(r - q - vol*vol))
    m2 = m2l + m2r

    volA = np.sqrt(np.log(m2/(m1*m1)) / T)

    d1 = (np.log(m1/k) + 0.5*volA*volA*T) / (volA*np.sqrt(T))
    d2 = d1 - volA*np.sqrt(T)

    call = np.exp(-r*T)*(m1*norm.cdf(d1) - k*norm.cdf(d2))
    put = np.exp(-r*T)*(k*norm.cdf(-d2) - m1*norm.cdf(-d1))
    return([call,put])

def Power(s,k,r,T,vol,q,n):
    '''
    Calculate the Black Scholes value of a traditional Power Call / Put option
    with a fixed strike
    Payoffs are of the form :
    C = max(S**n - K, 0)
    P = max(K - S**n, 0)

    Parameters
    ----------
    s : number of any type (int, float8, float64 etc.), numpy array of any type
        should the user wish to have a list of values output with varying s
        Spot value of underlying asset at current time, t
    k : number of any type (int, float8, float64 etc.), numpy array of any type
        should the user wish to have a list of values output with varying k
        Strike value of option, determined at initiation
    r : number of any type (int, float8, float64 etc.), numpy array of any type
        should the user wish to have a list of values output with varying r
        Risk free interest rate, implied constant till expiration
    T : number of any type (int, float8, float64 etc.), numpy array of any type
        should the user wish to have a list of values output with varying T
        Time till expiration for option
    vol : number of any type (int, float8, float64 etc.), numpy array of any type
        should the user wish to have a list of values output with varying vol
        Volatility of underlying, implied constant till expiration in Black
        Scholes model
    q : number of any type (int, float8, float64 etc.), numpy array of any type
        should the user wish to have a list of values output with varying q
        Continuous dividend payout, as a percentage
    n : number of any type (int, float8, float64 etc.)
        Power to which the underlying spot is raised at payoff

    All parameters can be individual values.
    Only one of these parameters can be a numpy.array (not including 'n'),
    otherwise there will be a dimension mismatch.

    Returns
    -------
    [call,put] : list of pair of float or numpy.array values
        Power call and put values, type depends on input value.
        If all input values are individual numbers, then output will be float.
        If one input value, other than 'n', is numpy.array, then output
        will be numpy.array.

    Examples
    --------
    >>> from qcfoptions.bsoptions import Power
    >>> s = 1
        k = 1
        r = 0.01
        T = 2
        vol = 0.25
        q = 0.015
        n = 2.5
    >>> Power(s,k,r,T,vol,q,n)
    [0.5888398346686554, 0.2369822536424761]
    >>> import numpy as np
    >>> s_array = np.array([0.5,1.0,1.5,2.0])
    >>> Power(s_array,k,r,T,vol,q,n)
    [array([0.00881893, 0.58883983, 2.7187886 , 6.51733355]),
    array([0.74551209, 0.23698225, 0.04539523, 0.00770309])]

    '''
    d1 = (np.log(s/np.power(k,1/n)) +
        (r - q + vol*vol*(n - 0.5))*T) / (vol*np.sqrt(T))
    d2 = d1 - n*vol*np.sqrt(T)

    option = np.exp(T*(n-1)*(r + 0.5*n*vol*vol))*np.power(s,n)
    strike = k*np.exp(-r*T)

    put = strike*norm.cdf(-d2) - option*norm.cdf(-d1)
    call = option*norm.cdf(d1) - strike*norm.cdf(d2)
    return([call,put])

def PowerStrike(s,k,r,T,vol,q,n):
    '''
    Calculate the Black Scholes value of Power Call / Put option with a
    fixed strike to the power n
    Payoffs are of the form :
    C = max(S**n - K**n, 0)
    P = max(K**n - S**n, 0)

    Parameters
    ----------
    s : number of any type (int, float8, float64 etc.), numpy array of any type
        should the user wish to have a list of values output with varying s
        Spot value of underlying asset at current time, t
    k : number of any type (int, float8, float64 etc.), numpy array of any type
        should the user wish to have a list of values output with varying k
        Strike value of option, determined at initiation
    r : number of any type (int, float8, float64 etc.), numpy array of any type
        should the user wish to have a list of values output with varying r
        Risk free interest rate, implied constant till expiration
    T : number of any type (int, float8, float64 etc.), numpy array of any type
        should the user wish to have a list of values output with varying T
        Time till expiration for option
    vol : number of any type (int, float8, float64 etc.), numpy array of any type
        should the user wish to have a list of values output with varying vol
        Volatility of underlying, implied constant till expiration in Black
        Scholes model
    q : number of any type (int, float8, float64 etc.), numpy array of any type
        should the user wish to have a list of values output with varying q
        Continuous dividend payout, as a percentage
    n : number of any type (int, float8, float64 etc.)
        Power to which the underlying spot is raised at payoff

    All parameters can be individual values.
    Only one of these parameters can be a numpy.array (not including 'n'),
    otherwise there will be a dimension mismatch.

    Returns
    -------
    [call,put] : list of pair of float or numpy.array values
        Power call and put values, type depends on input value.
        If all input values are individual numbers, then output will be float.
        If one input value, other than 'n', is numpy.array, then output
        will be numpy.array.

    Examples
    --------
    >>> from qcfoptions.bsoptions import PowerStrike
    >>> s = 1
        k = 1
        r = 0.01
        T = 2
        vol = 0.25
        q = 0.015
        n = 2.5
    >>> PowerStrike(s,k,r,T,vol,q,n)
    [0.5888398346686554, 0.2369822536424761]
    >>> import numpy as np
    >>> s_array = np.array([0.5,1.0,1.5,2.0])
    >>> PowerStrike(s_array,k,r,T,vol,q,n)
    [array([0.00881893, 0.58883983, 2.7187886 , 6.51733355]),
    array([0.74551209, 0.23698225, 0.04539523, 0.00770309])]

    '''
    d1 = (np.log(s/np.power(k,1/n)) +
        (r - q + vol*vol*(n - 0.5))*T) / (vol*np.sqrt(T))
    d2 = d1 - n*vol*np.sqrt(T)

    option = np.exp(T*(n-1)*(r + 0.5*n*vol*vol))*np.power(s,n)
    strike = np.power(k,n)*np.exp(-r*T)

    put = strike*norm.cdf(-d2) - option*norm.cdf(-d1)
    call = option*norm.cdf(d1) - strike*norm.cdf(d2)
    return([call,put])

def Margrabe(s,s2,T,vol,vol2,q,q2,corr):
    '''
    Calculate the Black Scholes value of the Margrabe Option
    Payoff is of the form :
    O = max(S_1 - S_2, 0)

    Parameters
    ----------
    s1 and s2 : number of any type (int, float8, float64 etc.), numpy array of
        any type should the user wish to have a list of values output with
        varying s
        Spot value of underlying assets 1 and 2 at current time, t
    T : number of any type (int, float8, float64 etc.), numpy array of any type
        should the user wish to have a list of values output with varying T
        Time till expiration for option
    vol1 and vol2 : number of any type (int, float8, float64 etc.), numpy array
        of any type should the user wish to have a list of values output with
        varying vol
        Volatility of underlying for assets 1 and 2, implied constant till
        expiration in Black Scholes model
    q1 and q2 : number of any type (int, float8, float64 etc.), numpy array of
        any type should the user wish to have a list of values output with
        varying q
        Continuous dividend payout for assets 1 and 2, as a percentage
    corr : number of any type (int, float8, float64 etc.), numpy array of any
        type should the user wish to have a list of values output with varying
        corr
        Correlation between the motion of the underlying (relationship between
        the Weiner process of asset 1 and 2)

    All parameters can be individual values.
    At most, only one pair of these parameters can be a numpy.array,
    otherwise there will be a dimension mismatch.

    Returns
    -------
    price : float or numpy.array value
        Margrabe price, type depends on input value.
        If all input values are individual numbers, then output will be float.
        If one pair of input values are a numpy.array,
        then output will be numpy.array.

    Examples
    --------
    >>> from qcfoptions.bsoptions import Margrabe
    >>> s1 = 1
        s2 = 1.2
        T = 2
        q1 = 0.1
        q2 = 0.05
        vol1 = 0.25
        vol2 = 0.15
        corr = 0.4
    >>> Margrabe(s1,s2,T,vol1,vol2,q1,q2,corr)
    0.0454498263366398
    >>> import numpy as np
    >>> s1_array = np.array([0.5,1.0,1.5,2.0])
        s2_array = np.array([0.5,1.0,1.5,2.0])
    >>> Margrabe(s1_array,s2_array,T,vol1,vol2,q1,q2,corr)
    array([0.04466118, 0.08932237, 0.13398355, 0.17864473])

    '''
    volMix = np.sqrt(vol*vol + vol2*vol2 - vol*vol2*corr)
    d1 = (np.log(s/s2) + (q2 - q + 0.5*(volMix**2))*T) / (volMix*np.sqrt(T))
    d2 = d1 - volMix*np.sqrt(T)

    option = np.exp(-q*T)*s*norm.cdf(d1)
    option2 = np.exp(-q2*T)*s2*norm.cdf(d2)

    price = option - option2
    return(price)

def Lookback(s,M,r,T,vol,q):
    '''
    Calculate the Black Scholes value of floating strike
    Lookback Call / Put option
    Payoffs are of the form :
    C = S_T - min(m,m_T)
    P = max(M,M_T) - S_T
    where 'm' is the current minimum, or starting strike at initiation, and
    'm_T' is the minimum over the remaining life of the option
    similarly, 'M' is the current maximum, or starting strike at initiation, and
    'M_T' is the maximum over the remaining life of the option

    Parameters
    ----------
    s : number of any type (int, float8, float64 etc.), numpy array of any type
        should the user wish to have a list of values output with varying s
        Spot value of underlying asset at current time, t
    M : number of any type (int, float8, float64 etc.), numpy array of any type
        should the user wish to have a list of values output with varying k
        Strike value of option, determined by minimum value of underlying
        over the life of the option
    r : number of any type (int, float8, float64 etc.), numpy array of any type
        should the user wish to have a list of values output with varying r
        Risk free interest rate, implied constant till expiration
    T : number of any type (int, float8, float64 etc.), numpy array of any type
        should the user wish to have a list of values output with varying T
        Time till expiration for option
    vol : number of any type (int, float8, float64 etc.), numpy array of any type
        should the user wish to have a list of values output with varying vol
        Volatility of underlying, implied constant till expiration in Black
        Scholes model
    q : number of any type (int, float8, float64 etc.), numpy array of any type
        should the user wish to have a list of values output with varying q
        Continuous dividend payout, as a percentage

    All parameters can be individual values.
    Only one of these parameters can be a numpy.array, otherwise there will be
    a dimension mismatch.

    Returns
    -------
    [call,put] : list of pair of float or numpy.array values
        Lookback call and put values, type depends on input value.
        If all input values are individual numbers, then output will be float.
        If one input value is numpy.array, then output will be numpy.array.

    Examples
    --------
    >>> from qcfoptions.bsoptions import Lookback
    >>> s = 1
        k = 1
        r = 0.015
        T = 2
        vol = 0.25
        q = 0.01
    >>> Lookback(s,M,r,T,vol,q)
    [0.2509969952730551, 0.30220097900385257]
    >>> import numpy as np
    >>> s_array = np.array([0.5,1.0,1.5,2.0])
    >>> Lookback(s_array,M,r,T,vol,q)
    [array([0.33036841, 0.250997  , 0.54619226, 0.99797004]),
    array([0.48550323, 0.30220098, 0.70926038, 1.40333811])]

    '''
    B = 2*(r - q) / (vol*vol)
    x = (np.log(s/M) + (r - q - 0.5*vol*vol)*T) / (vol*np.sqrt(T))
    y = (-np.log(s/M) - (r - q + 0.5*vol*vol)*T) / (vol*np.sqrt(T))

    option = s*np.exp(-q*T)
    minimum = M*np.exp(-r*T)
    left = np.exp(-r*T)*np.power(s/M,-B)
    right = np.exp(-q*T)

    call = (option*norm.cdf(x + vol*np.sqrt(T)) -
        minimum*norm.cdf(x) +
        (s/B)*(left*norm.cdf(y + B*vol*np.sqrt(T)) -
            right*norm.cdf(y)))
    put = (-option*norm.cdf(-x - vol*np.sqrt(T)) +
        minimum*norm.cdf(-x) -
        (s/B)*(left*norm.cdf(-y - B*vol*np.sqrt(T)) -
            right*norm.cdf(-y)))
    return([call,put])




# Austin Griffith
# Setup File

from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    # name of package
    # setup using
    # $ pip install qcfoptions
    name = 'qcfoptions',

    version = '0.1.2.1',
    description = 'Option calculator',

    # uses readme file to give long description
    long_description = long_description,
    long_description_content_type = 'text/markdown',

    url = 'https://github.com/austingriffith94/qcfoptions',
    author = 'Austin Griffith',
    license = 'MIT',

    classifiers = [
        'Development Status :: 2 - Pre-Alpha',
        'License :: OSI Approved :: MIT License',
        # began with python 3.6.5
        'Programming Language :: Python :: 3.6',
    ],

    python_requires = '~=3.6',
    packages = ['qcfoptions'],
    install_requires = ['numpy','scipy'],
    )



# Austin Griffith
# Simulation

import numpy as np
import scipy.stats as sctats
import time

# simulation payoffs
def EuroSim(S,k,r,T):
    '''
    Use simulated underlying to determine the price of a European
    Call / Put option
    Payoffs are of the form :
    C = max(S - K, 0)
    P = max(K - S, 0)

    Parameters
    ----------
    S : numpy.array
        Simulated stock price, want to be of the form such that the first row
        is the initial stock price, with subsequent rows representing an
        additional time step increase, and each column is a simulated path of
        the asset
    k : number of any type (int, float8, float64 etc.)
        Strike value of option, determined at initiation
    r : number of any type (int, float8, float64 etc.)
        Risk free interest rate, implied constant till expiration
    T : number of any type (int, float8, float64 etc.)
        Time till expiration for option

    Returns
    -------
    [[call,put],[callMotion,putMotion]] : list of pair of lists, first of
        floats, second of one-dimensional numpy.array's
        First list is the call and put price, determined by the average
        of the simulated stock payoffs
        Second list is the call and put simulated paths payoffs at expiration,
        NOT discounted

    * the accuracy of pricing is dependent on the number of time steps and
    simulated paths chosen for the underlying stochastic motion

    Examples
    --------
    >>> from simulation import EuroSim
    >>> import numpy as np
    >>> s0 = 1
        r = 0.015
        T = 0.5
        k = 0.8
    >>> S = np.array([[ 1.        ,  1.        ,  1.        ,  1.        ,  1.        ],
                   [ 0.92248705,  1.08050869,  0.92248705,  1.08050869,  0.92248705],
                   [ 0.85098236,  0.99675528,  0.85098236,  0.99675528,  0.99675528],
                   [ 0.91949383,  0.91949383,  0.91949383,  1.07700274,  0.91949383],
                   [ 0.99352108,  0.99352108,  0.84822115,  1.16371082,  0.99352108],
                   [ 0.91651033,  1.07350816,  0.78247303,  1.07350816,  1.07350816]])
    >>> a = EuroSim(S,k,r,T)
    >>> print(a[0])
        print(a[1][0])
        print(a[1][1])
        [0.1860066674534242, 0.0034792018881946852]
        [ 0.11651033  0.27350816  0.          0.27350816  0.27350816]
        [ 0.          0.          0.01752697  0.          0.        ]

    '''
    callMotion = (S[-1] - k).clip(0)
    putMotion = (k - S[-1]).clip(0)

    call = np.exp(-r*T)*np.average(callMotion)
    put = np.exp(-r*T)*np.average(putMotion)
    return([[call,put],[callMotion,putMotion]])

def AsianGeoFixSim(S,k,r,T):
    '''
    Use simulated underlying to determine the price of an Asian Geometric
    Average Call / Put option with a fixed strike price
    Payoffs are of the form :
    C = max(AVG_geo - K, 0)
    P = max(K - AVG_geo, 0)

    Parameters
    ----------
    S : numpy.array
        Simulated stock price, want to be of the form such that the first row
        is the initial stock price, with subsequent rows representing an
        additional time step increase, and each column is a simulated path of
        the asset
    k : number of any type (int, float8, float64 etc.)
        Strike value of option, determined at initiation
    r : number of any type (int, float8, float64 etc.)
        Risk free interest rate, implied constant till expiration
    T : number of any type (int, float8, float64 etc.)
        Time till expiration for option

    Returns
    -------
    [[call,put],[callMotion,putMotion]] : list of pair of lists, first of
        floats, second of one-dimensional numpy.array's
        First list is the call and put price, determined by the average
        of the simulated stock payoffs
        Second list is the call and put simulated paths payoffs at expiration,
        NOT discounted

    * the accuracy of pricing is dependent on the number of time steps and
    simulated paths chosen for the underlying stochastic motion

    Examples
    --------
    >>> from simulation import AsianGeoFixSim
    >>> import numpy as np
    >>> s0 = 1
        r = 0.015
        T = 0.5
        k = 0.8
    >>> S = np.array([[ 1.        ,  1.        ,  1.        ,  1.        ,  1.        ],
                   [ 0.92248705,  1.08050869,  0.92248705,  1.08050869,  0.92248705],
                   [ 0.85098236,  0.99675528,  0.85098236,  0.99675528,  0.99675528],
                   [ 0.91949383,  0.91949383,  0.91949383,  1.07700274,  0.91949383],
                   [ 0.99352108,  0.99352108,  0.84822115,  1.16371082,  0.99352108],
                   [ 0.91651033,  1.07350816,  0.78247303,  1.07350816,  1.07350816]])
    >>> a = AsianGeoFixSim(S,k,r,T)
    >>> print(a[0])
        print(a[1][0])
        print(a[1][1])
        [0.17326664943627884, 0.0]
        [ 0.1324467   0.20915531  0.08457506  0.26376902  0.18290908]
        [ 0.  0.  0.  0.  0.]

    '''
    avg = sctats.gmean(S,axis=0)
    callMotion = (avg - k).clip(0)
    putMotion = (k - avg).clip(0)

    call = np.exp(-r*T)*np.average(callMotion)
    put = np.exp(-r*T)*np.average(putMotion)
    return([[call,put],[callMotion,putMotion]])

def AsianGeoFloatSim(S,m,r,T):
    '''
    Use simulated underlying to determine the price of an Asian Geometric
    Average Call / Put option with a floating strike price
    Payoffs are of the form :
    C = max(S - m*AVG_geo, 0)
    P = max(m*AVG_geo - S, 0)

    Parameters
    ----------
    S : numpy.array
        Simulated stock price, want to be of the form such that the first row
        is the initial stock price, with subsequent rows representing an
        additional time step increase, and each column is a simulated path of
        the asset
    m : number of any type (int, float8, float64 etc.)
        Strike value scaler of option, determined at initiation
    r : number of any type (int, float8, float64 etc.)
        Risk free interest rate, implied constant till expiration
    T : number of any type (int, float8, float64 etc.)
        Time till expiration for option

    Returns
    -------
    [[call,put],[callMotion,putMotion]] : list of pair of lists, first of
        floats, second of one-dimensional numpy.array's
        First list is the call and put price, determined by the average
        of the simulated stock payoffs
        Second list is the call and put simulated paths payoffs at expiration,
        NOT discounted

    * the accuracy of pricing is dependent on the number of time steps and
    simulated paths chosen for the underlying stochastic motion

    Examples
    --------
    >>> from simulation import AsianGeoFloatSim
    >>> import numpy as np
    >>> s0 = 1
        r = 0.015
        T = 0.5
        m = 0.8
    >>> S = np.array([[ 1.        ,  1.        ,  1.        ,  1.        ,  1.        ],
                   [ 0.92248705,  1.08050869,  0.92248705,  1.08050869,  0.92248705],
                   [ 0.85098236,  0.99675528,  0.85098236,  0.99675528,  0.99675528],
                   [ 0.91949383,  0.91949383,  0.91949383,  1.07700274,  0.91949383],
                   [ 0.99352108,  0.99352108,  0.84822115,  1.16371082,  0.99352108],
                   [ 0.91651033,  1.07350816,  0.78247303,  1.07350816,  1.07350816]])
    >>> a = AsianGeoFloatSim(S,k,r,T)
    >>> print(a[0])
        print(a[1][0])
        print(a[1][1])
        [0.20271863478726854, 0.0]
        [ 0.17055297  0.26618391  0.07481299  0.22249294  0.2871809 ]
        [ 0.  0.  0.  0.  0.]

    '''
    avg = sctats.gmean(S,axis=0)
    callMotion = (S[-1] - m*avg).clip(0)
    putMotion = (m*avg - S[-1]).clip(0)

    call = np.exp(-r*T)*np.average(callMotion)
    put = np.exp(-r*T)*np.average(putMotion)
    return([[call,put],[callMotion,putMotion]])

def AsianArithFixSim(S,k,r,T):
    '''
    Use simulated underlying to determine the price of an Asian Arithmetic
    Average Call / Put option with a fixed strike price
    Payoffs are of the form :
    C = max(AVG_arithmetic - K, 0)
    P = max(K - AVG_arithmetic, 0)

    Parameters
    ----------
    S : numpy.array
        Simulated stock price, want to be of the form such that the first row
        is the initial stock price, with subsequent rows representing an
        additional time step increase, and each column is a simulated path of
        the asset
    k : number of any type (int, float8, float64 etc.)
        Strike value of option, determined at initiation
    r : number of any type (int, float8, float64 etc.)
        Risk free interest rate, implied constant till expiration
    T : number of any type (int, float8, float64 etc.)
        Time till expiration for option

    Returns
    -------
    [[call,put],[callMotion,putMotion]] : list of pair of lists, first of
        floats, second of one-dimensional numpy.array's
        First list is the call and put price, determined by the average
        of the simulated stock payoffs
        Second list is the call and put simulated paths payoffs at expiration,
        NOT discounted

    * the accuracy of pricing is dependent on the number of time steps and
    simulated paths chosen for the underlying stochastic motion

    Examples
    --------
    >>> from simulation import AsianArithFixSim
    >>> import numpy as np
    >>> s0 = 1
        r = 0.015
        T = 0.5
        k = 0.8
    >>> S = np.array([[ 1.        ,  1.        ,  1.        ,  1.        ,  1.        ],
                   [ 0.92248705,  1.08050869,  0.92248705,  1.08050869,  0.92248705],
                   [ 0.85098236,  0.99675528,  0.85098236,  0.99675528,  0.99675528],
                   [ 0.91949383,  0.91949383,  0.91949383,  1.07700274,  0.91949383],
                   [ 0.99352108,  0.99352108,  0.84822115,  1.16371082,  0.99352108],
                   [ 0.91651033,  1.07350816,  0.78247303,  1.07350816,  1.07350816]])
    >>> a = AsianArithFixSim(S,k,r,T)
    >>> print(a[0])
        print(a[1][0])
        print(a[1][1])
        [0.17493936228974066, 0.0]
        [ 0.13383244  0.21063117  0.08727624  0.26524762  0.18429423]
        [ 0.  0.  0.  0.  0.]

    '''
    avg = np.average(S,axis=0)
    callMotion = (avg - k).clip(0)
    putMotion = (k - avg).clip(0)

    call = np.exp(-r*T)*np.average(callMotion)
    put = np.exp(-r*T)*np.average(putMotion)
    return([[call,put],[callMotion,putMotion]])

def AsianArithFloatSim(S,m,r,T):
    '''
    Use simulated underlying to determine the price of an Asian Arithmetic
    Average Call / Put option with a floating strike price
    Payoffs are of the form :
    C = max(S - m*AVG_arithmetic, 0)
    P = max(m*AVG_arithmetic - S, 0)

    Parameters
    ----------
    S : numpy.array
        Simulated stock price, want to be of the form such that the first row
        is the initial stock price, with subsequent rows representing an
        additional time step increase, and each column is a simulated path of
        the asset
    m : number of any type (int, float8, float64 etc.)
        Strike value scaler of option, determined at initiation
    r : number of any type (int, float8, float64 etc.)
        Risk free interest rate, implied constant till expiration
    T : number of any type (int, float8, float64 etc.)
        Time till expiration for option

    Returns
    -------
    [[call,put],[callMotion,putMotion]] : list of pair of lists, first of
        floats, second of one-dimensional numpy.array's
        First list is the call and put price, determined by the average
        of the simulated stock payoffs
        Second list is the call and put simulated paths payoffs at expiration,
        NOT discounted

    * the accuracy of pricing is dependent on the number of time steps and
    simulated paths chosen for the underlying stochastic motion

    Examples
    --------
    >>> from simulation import AsianArithFloatSim
    >>> import numpy as np
    >>> s0 = 1
        r = 0.015
        T = 0.5
        m = 0.8
    >>> S = np.array([[ 1.        ,  1.        ,  1.        ,  1.        ,  1.        ],
                   [ 0.92248705,  1.08050869,  0.92248705,  1.08050869,  0.92248705],
                   [ 0.85098236,  0.99675528,  0.85098236,  0.99675528,  0.99675528],
                   [ 0.91949383,  0.91949383,  0.91949383,  1.07700274,  0.91949383],
                   [ 0.99352108,  0.99352108,  0.84822115,  1.16371082,  0.99352108],
                   [ 0.91651033,  1.07350816,  0.78247303,  1.07350816,  1.07350816]])
    >>> a = AsianArithFloatSim(S,k,r,T)
    >>> print(a[0])
        print(a[1][0])
        print(a[1][1])
        [0.20138046450449909, 0.0]
        [ 0.16944438  0.26500322  0.07265204  0.22131007  0.28607277]
        [ 0.  0.  0.  0.  0.]

    '''
    avg = np.average(S,axis=0)
    callMotion = (S[-1] - m*avg).clip(0)
    putMotion = (m*avg - S[-1]).clip(0)

    # class
    call = np.exp(-r*T)*np.average(callMotion)
    put = np.exp(-r*T)*np.average(putMotion)
    return([[call,put],[callMotion,putMotion]])

def PowerSim(S,k,r,T,n):
    '''
    Use simulated underlying to determine the price of a Power Call / Put
    option with a fixed strike price
    Payoffs are of the form :
    C = max(S**n - K, 0)
    P = max(K - S**n, 0)

    Parameters
    ----------
    S : numpy.array
        Simulated stock price, want to be of the form such that the first row
        is the initial stock price, with subsequent rows representing an
        additional time step increase, and each column is a simulated path of
        the asset
    k : number of any type (int, float8, float64 etc.)
        Strike value of option, determined at initiation
    r : number of any type (int, float8, float64 etc.)
        Risk free interest rate, implied constant till expiration
    T : number of any type (int, float8, float64 etc.)
        Time till expiration for option
    n : number of any type (int, float8, float64 etc.)
        Power the underlying is raised to at expiration

    Returns
    -------
    [[call,put],[callMotion,putMotion]] : list of pair of lists, first of
        floats, second of one-dimensional numpy.array's
        First list is the call and put price, determined by the average
        of the simulated stock payoffs
        Second list is the call and put simulated paths payoffs at expiration,
        NOT discounted

    * the accuracy of pricing is dependent on the number of time steps and
    simulated paths chosen for the underlying stochastic motion

    Examples
    --------
    >>> from simulation import PowerSim
    >>> import numpy as np
    >>> s0 = 1
        r = 0.015
        T = 0.5
        k = 0.8
        n = 2.5
    >>> S = np.array([[ 1.        ,  1.        ,  1.        ,  1.        ,  1.        ],
                   [ 0.92248705,  1.08050869,  0.92248705,  1.08050869,  0.92248705],
                   [ 0.85098236,  0.99675528,  0.85098236,  0.99675528,  0.99675528],
                   [ 0.91949383,  0.91949383,  0.91949383,  1.07700274,  0.91949383],
                   [ 0.99352108,  0.99352108,  0.84822115,  1.16371082,  0.99352108],
                   [ 0.91651033,  1.07350816,  0.78247303,  1.07350816,  1.07350816]])
    >>> a = PowerSim(S,k,r,T)
    >>> print(a[0])
        print(a[1][0])
        print(a[1][1])
        [0.23547457653967713, 0.051295140170868392]
        [ 0.00416175  0.39402488  0.          0.39402488  0.39402488]
        [ 0.         0.         0.2584065  0.         0.       ]

    '''
    power = np.power(S[-1],n)
    callMotion = (power - k).clip(0)
    putMotion = (k - power).clip(0)

    call = np.exp(-r*T)*np.average(callMotion)
    put = np.exp(-r*T)*np.average(putMotion)
    return([[call,put],[callMotion,putMotion]])

def PowerStrikeSim(S,k,r,T,n):
    '''
    Use simulated underlying to determine the price of a Power Call / Put
    option with a fixed strike price
    Payoffs are of the form :
    C = max(S**n - K**n, 0)
    P = max(K**n - S**n, 0)

    Parameters
    ----------
    S : numpy.array
        Simulated stock price, want to be of the form such that the first row
        is the initial stock price, with subsequent rows representing an
        additional time step increase, and each column is a simulated path of
        the asset
    k : number of any type (int, float8, float64 etc.)
        Strike value of option, determined at initiation
    r : number of any type (int, float8, float64 etc.)
        Risk free interest rate, implied constant till expiration
    T : number of any type (int, float8, float64 etc.)
        Time till expiration for option
    n : number of any type (int, float8, float64 etc.)
        Power the underlying and strike are raised to at expiration

    Returns
    -------
    [[call,put],[callMotion,putMotion]] : list of pair of lists, first of
        floats, second of one-dimensional numpy.array's
        First list is the call and put price, determined by the average
        of the simulated stock payoffs
        Second list is the call and put simulated paths payoffs at expiration,
        NOT discounted

    * the accuracy of pricing is dependent on the number of time steps and
    simulated paths chosen for the underlying stochastic motion

    Examples
    --------
    >>> from simulation import PowerStrikeSim
    >>> import numpy as np
    >>> s0 = 1
        r = 0.015
        T = 0.5
        k = 0.8
        n = 2.5
    >>> S = np.array([[ 1.        ,  1.        ,  1.        ,  1.        ,  1.        ],
                   [ 0.92248705,  1.08050869,  0.92248705,  1.08050869,  0.92248705],
                   [ 0.85098236,  0.99675528,  0.85098236,  0.99675528,  0.99675528],
                   [ 0.91949383,  0.91949383,  0.91949383,  1.07700274,  0.91949383],
                   [ 0.99352108,  0.99352108,  0.84822115,  1.16371082,  0.99352108],
                   [ 0.91651033,  1.07350816,  0.78247303,  1.07350816,  1.07350816]])
    >>> a = PowerStrikeSim(S,k,r,T)
    >>> print(a[0])
        print(a[1][0])
        print(a[1][1])
        [0.41616756263295357, 0.0061218936475492848]
        [ 0.23172835  0.62159147  0.          0.62159147  0.62159147]
        [ 0.         0.         0.0308399  0.         0.       ]

    '''
    powerS = np.power(S[-1],n)
    callMotion = (powerS - k**n).clip(0)
    putMotion = (k**n - powerS).clip(0)

    call = np.exp(-r*T)*np.average(callMotion)
    put = np.exp(-r*T)*np.average(putMotion)
    return([[call,put],[callMotion,putMotion]])

def AvgBarrierSim(S,Z,r,timeMatrix):
    '''
    Use simulated underlying to determine the price of an Average Barrier
    option, where the payoff is the arithmetic average of the underlying over
    the time t (t is time when the option hits the barrier), or
    T (time to expiration) if the barrier is never hit
    Payoff is of the form :
    P = AVG_arithmetic_t

    Parameters
    ----------
    S : numpy.array
        Simulated stock price, want to be of the form such that the first row
        is the initial stock price, with subsequent rows representing an
        additional time step increase, and each column is a simulated path of
        the asset
    Z : number of any type (int, float8, float64 etc.)
        Barrier value of option, determined at initiation
    r : number of any type (int, float8, float64 etc.)
        Risk free interest rate, implied constant till expiration
    timeMatrix : numpy.matrix
        Matrix of time intervals, of the same dimensions as the S matrix,
        should have 'paths' number of columns, and rows that iterate between 0
        and T by dt

    Returns
    -------
    [price,payoffMotion] : list, first is float, second is
        one-dimensional numpy.array
        price, is estimated price of the option, determined by the average
        of the simulated stock payoffs
        payoffMotion is the simulated paths payoffs at expiration,
        NOT discounted

    * the accuracy of pricing is dependent on the number of time steps and
    simulated paths chosen for the underlying stochastic motion
    ** if the barrier is equal to the initial spot price, the price and
    payoffMotion will both be equal to the spot price since underlying hits the
    barrier at initiation

    Examples
    --------
    >>> from simulation import AvgBarrierSim
    >>> import numpy as np
    >>> s0 = 1
        r = 0.015
        T = 0.5
        z = 1.1
    >>> S = np.array([[ 1.        ,  1.        ,  1.        ,  1.        ,  1.        ],
                   [ 0.92248705,  1.08050869,  0.92248705,  1.08050869,  0.92248705],
                   [ 0.85098236,  0.99675528,  0.85098236,  0.99675528,  0.99675528],
                   [ 0.91949383,  0.91949383,  0.91949383,  1.07700274,  0.91949383],
                   [ 0.99352108,  0.99352108,  0.84822115,  1.16371082,  0.99352108],
                   [ 0.91651033,  1.07350816,  0.78247303,  1.07350816,  1.07350816]])
    >>> timeMatrix = timeMatrix = np.matrix([[ 0. ,  0. ,  0. ,  0. ,  0. ],
                                            [ 0.1,  0.1,  0.1,  0.1,  0.1],
                                            [ 0.2,  0.2,  0.2,  0.2,  0.2],
                                            [ 0.3,  0.3,  0.3,  0.3,  0.3],
                                            [ 0.4,  0.4,  0.4,  0.4,  0.4],
                                            [ 0.5,  0.5,  0.5,  0.5,  0.5]])
    >>> a = AvgBarrierSim(S,z,r,timeMatrix)
    >>> print(a[0])
        print(a[1])
        0.964284902938
        [ 0.93383244  1.01063117  0.88727624  1.03856668  0.98429423]

    '''
    s0 = S[0][0]
    if s0 < Z: # below
        hitBarrier = np.cumprod(S < Z,axis=0)
    elif s0 > Z: # above
        hitBarrier = np.cumprod(S > Z,axis=0)
    else: # on barrier
        price = s0
        payoffMotion = S[0]
        return([price,payoffMotion])

    paymentTime = np.array(np.max(np.multiply(timeMatrix,hitBarrier),axis=0))
    payoffMotion = np.sum(np.multiply(hitBarrier,S),axis=0) / np.sum(hitBarrier,axis=0)
    price = np.average(np.exp(-r*paymentTime)*payoffMotion)
    return([price,payoffMotion])

def NoTouchSingleSim(S,Z,r,T,payoutScale):
    '''
    Use simulated underlying to determine the price of a No Touch Binary
    option, with a single direction having a barrier. If the barrier is hit
    prior to expiration, their is no payoff. If it isn't, the payoff has a
    scale prior to initiation.
    Payoff is of the form :
    P = (money down * payoffScale, 0)_Z

    Parameters
    ----------
    S : numpy.array
        Simulated stock price, want to be of the form such that the first row
        is the initial stock price, with subsequent rows representing an
        additional time step increase, and each column is a simulated path of
        the asset
    Z : number of any type (int, float8, float64 etc.)
        Barrier value of option, determined at initiation
    r : number of any type (int, float8, float64 etc.)
        Risk free interest rate, implied constant till expiration
    T : number of any type (int, float8, float64 etc.)
        Time till expiration for option
    payoffScale : number of any type (int, float8, float64 etc.)
        Scale value of payoff, should be a percentage (e.g. 20% payoff should
        be 0.2 when input)

    Returns
    -------
    [price,payoffMotion] : list, first is float, second is
        one-dimensional numpy.array
        price, is estimated return on option per dollar put down,
        determined by the average of the simulated stock payoffs
        payoffMotion is the simulated paths payoffs at expiration,
        NOT discounted

    * the accuracy of pricing is dependent on the number of time steps and
    simulated paths chosen for the underlying stochastic motion
    ** if the barrier is equal to the initial spot price, the price and
    payoffMotion will both be 0 since underlying hits the barrier at initiation

    Examples
    --------
    >>> from simulation import NoTouchSingleSim
    >>> import numpy as np
    >>> s0 = 1
        r = 0.015
        T = 0.5
        z = 1.1
        payoffScale = 0.5
    >>> S = np.array([[ 1.        ,  1.        ,  1.        ,  1.        ,  1.        ],
                   [ 0.92248705,  1.08050869,  0.92248705,  1.08050869,  0.92248705],
                   [ 0.85098236,  0.99675528,  0.85098236,  0.99675528,  0.99675528],
                   [ 0.91949383,  0.91949383,  0.91949383,  1.07700274,  0.91949383],
                   [ 0.99352108,  0.99352108,  0.84822115,  1.16371082,  0.99352108],
                   [ 0.91651033,  1.07350816,  0.78247303,  1.07350816,  1.07350816]])
    >>> a = NoTouchSingleSim(S,z,r,T,payoffScale)
    >>> print(a[0])
        print(a[1])
        1.19103366578
        [ 1.5  1.5  1.5  0.   1.5]

    '''
    s0 = S[0][0]
    if s0 < Z: # below
        hitBarrier = np.cumprod(S < Z,axis=0)
    elif s0 > Z: # above
        hitBarrier = np.cumprod(S > Z,axis=0)
    else: # on barrier
        price = 0.0
        payoffMotion = S[0]*0.0
        return([price,payoffMotion])

    payoffMotion = (1+payoutScale)*hitBarrier[-1]
    price = np.average(np.exp(-r*T)*payoffMotion)
    return([price,payoffMotion])

def NoTouchDoubleSim(S,Z1,Z2,r,T,payoutScale):
    '''
    Use simulated underlying to determine the price of a No Touch Binary
    option, with both directions having a barrier. If either barrier is hit
    prior to expiration, their is no payoff. If they aren't hit, the payoff
    has a scale prior to initiation.
    Payoff is of the form :
    P = (money down * payoffScale, 0)_Z1,Z2

    Parameters
    ----------
    S : numpy.array
        Simulated stock price, want to be of the form such that the first row
        is the initial stock price, with subsequent rows representing an
        additional time step increase, and each column is a simulated path of
        the asset
    Z1 : number of any type (int, float8, float64 etc.)
        First barrier value of option, determined at initiation
    Z2 : number of any type (int, float8, float64 etc.)
        Second barrier value of option, determined at initiation
    r : number of any type (int, float8, float64 etc.)
        Risk free interest rate, implied constant till expiration
    T : number of any type (int, float8, float64 etc.)
        Time till expiration for option
    payoffScale : number of any type (int, float8, float64 etc.)
        Scale value of payoff, should be a percentage (e.g. 20% payoff should
        be 0.2 when input)

    Returns
    -------
    [price,payoffMotion] : list, first is float, second is
        one-dimensional numpy.array
        price, is estimated return on option per dollar put down,
        determined by the average of the simulated stock payoffs
        payoffMotion is the simulated paths payoffs at expiration,
        NOT discounted

    * the accuracy of pricing is dependent on the number of time steps and
    simulated paths chosen for the underlying stochastic motion
    ** if either barrier is equal to the initial spot price, the price and
    payoffMotion will both be 0 since underlying hits the barrier at initiation
    *** if spot is not between Z1 and Z2, then output error, since two
    barriers will be redundant

    Examples
    --------
    >>> from simulation import NoTouchDoubleSim
    >>> import numpy as np
    >>> s0 = 1
        r = 0.015
        T = 0.5
        z1 = 1.1
        z2 = 0.9
        payoffScale = 0.5
    >>> S = np.array([[ 1.        ,  1.        ,  1.        ,  1.        ,  1.        ],
                   [ 0.92248705,  1.08050869,  0.92248705,  1.08050869,  0.92248705],
                   [ 0.85098236,  0.99675528,  0.85098236,  0.99675528,  0.99675528],
                   [ 0.91949383,  0.91949383,  0.91949383,  1.07700274,  0.91949383],
                   [ 0.99352108,  0.99352108,  0.84822115,  1.16371082,  0.99352108],
                   [ 0.91651033,  1.07350816,  0.78247303,  1.07350816,  1.07350816]])
    >>> a = NoTouchDoubleSim(S,z,z2,r,T,payoffScale)
    >>> print(a[0])
        print(a[1])
        0.595516832891
        [ 0.   1.5  0.   0.   1.5]
    >>> z2 = 1.5
    >>> NoTouchDoubleSim(S,z,z2,r,T,payoffScale)
        Error : s0 outside barriers, use NoTouchSingle instead

    '''
    s0 = S[0][0]
    if s0 < Z1 and s0 > Z2:
        hitBarrier1 = np.cumprod(S < Z1,axis=0)
        hitBarrier2 = np.cumprod(S > Z2,axis=0)
    elif s0 > Z1 and s0 < Z2:
        hitBarrier1 = np.cumprod(S > Z1,axis=0)
        hitBarrier2 = np.cumprod(S < Z2,axis=0)
    elif s0 == Z1 or s0 == Z2:
        price = 0.0
        payoffMotion = S[0]*0.0
        return([price,payoffMotion])
    else:
        print('Error : s0 outside barriers, use NoTouchSingle instead')
        return

    hitBarrier = np.multiply(hitBarrier1,hitBarrier2)
    payoffMotion = (1+payoutScale)*hitBarrier[-1]
    price = np.average(np.exp(-r*T)*payoffMotion)
    return([price,payoffMotion])

def CashOrNothingSim(S,Z,r,T,payout):
    '''
    Use simulated underlying to determine the price of a Cash-or-Nothing
    option, with a single direction having a barrier. If the barrier is hit
    prior to expiration, their is no payoff. If it isn't, the payoff is a
    value determined prior to initiation.
    Payoff is of the form :
    P = (payout, 0)_Z

    Parameters
    ----------
    S : numpy.array
        Simulated stock price, want to be of the form such that the first row
        is the initial stock price, with subsequent rows representing an
        additional time step increase, and each column is a simulated path of
        the asset
    Z : number of any type (int, float8, float64 etc.)
        Barrier value of option, determined at initiation
    r : number of any type (int, float8, float64 etc.)
        Risk free interest rate, implied constant till expiration
    T : number of any type (int, float8, float64 etc.)
        Time till expiration for option
    payout : number of any type (int, float8, float64 etc.)
        Payout of option, fixed value paid out if the barrier isn't hit by the
        underlying over the life of the option

    Returns
    -------
    [price,payoffMotion] : list, first is float, second is
        one-dimensional numpy.array
        price, is estimated price of the option, determined by the average
        of the simulated stock payoffs
        payoffMotion is the simulated paths payoffs at expiration,
        NOT discounted

    * the accuracy of pricing is dependent on the number of time steps and
    simulated paths chosen for the underlying stochastic motion
    ** if the barrier is equal to the initial spot price, the price and
    payoffMotion will both be 0 since underlying hits the barrier at initiation

    Examples
    --------
    >>> from simulation import CashOrNothingSim
    >>> import numpy as np
    >>> s0 = 1
        r = 0.015
        T = 0.5
        z = 1.1
        payout = 100
    >>> S = np.array([[ 1.        ,  1.        ,  1.        ,  1.        ,  1.        ],
                   [ 0.92248705,  1.08050869,  0.92248705,  1.08050869,  0.92248705],
                   [ 0.85098236,  0.99675528,  0.85098236,  0.99675528,  0.99675528],
                   [ 0.91949383,  0.91949383,  0.91949383,  1.07700274,  0.91949383],
                   [ 0.99352108,  0.99352108,  0.84822115,  1.16371082,  0.99352108],
                   [ 0.91651033,  1.07350816,  0.78247303,  1.07350816,  1.07350816]])
    >>> a = CashOrNothingSim(S,z,r,T,payout)
    >>> print(a[0])
        print(a[1])
        79.4022443855
        [100 100 100   0 100]

    '''
    s0 = S[0][0]
    if s0 < Z: # below
        hitBarrier = np.cumprod(S < Z,axis=0)
    elif s0 > Z: # above
        hitBarrier = np.cumprod(S > Z,axis=0)
    else: # on barrier
        price = 0.0
        payoffMotion = S[0]*0.0
        return([price,payoffMotion])

    payoffMotion = hitBarrier[-1]*payout
    price = np.average(np.exp(-r*T)*payoffMotion)
    return([price,payoffMotion])

def SimpleSim(s0,r,T,vol,dt,paths):
    '''
    Simulate the motion of an underlying stock that follows a
    standard Weiner process for T/dt steps over a specified number of paths

    Parameters
    ----------
    s0 : number of any type (int, float8, float64 etc.)
        Spot value of underlying assets at current time, t
    r : number of any type (int, float8, float64 etc.)
        Risk free interest rate value
    T : number of any type (int, float8, float64 etc.)
        Time till expiration for option
    vol : number of any type (int, float8, float64 etc.)
        Volatility of underlying, implied constant till
        expiration in simple model
    dt : number of any type (int, float8, float64 etc.)
        Time interval used in simulation, used for number
        of stepsstock has, and motion of Weiner process
    paths : number of type 'int'
        Number of stocks simulated, higher number of paths
        leads to greater accuracy in calculated price.

    * numpy matrices will begin to break once a large enough path is chosen
    ** similarly, if too small of a dt is chosen, the numpy matrices
    will begin to break

    Returns
    -------
    S : numpy.array
        A (T/dt) x paths array that holds the simulated
        stock price values

        of the form:
        [[s_0 s_0 ... s_0],
        [s_11 s_12 ... s_1paths]
        ...
        [s_(T/dt)1 s_(T/dt)2 ... s_(T/dt)paths]]

    Examples
    --------
    >>> from qcfoptions.simulation import SimpleSim
    >>> s0 = 1
        r = 0.015
        T = 2
        vol = 0.25
        dt = 0.001
        paths = 1000
    >>> SimpleSim(s0,r,T,vol,dt,paths)
    array([[ 1.        ,  1.        ,  1.        , ...,  1.        ,
         1.        ,  1.        ],
       [ 1.00792065,  1.00792065,  1.00792065, ...,  1.00792065,
         0.99210935,  1.00792065],
       [ 1.01590403,  0.9999675 ,  1.01590403, ...,  1.01590403,
         0.9999675 ,  1.01590403],
       ...,
       [ 0.83965012,  0.96805391,  0.86662648, ...,  1.34928073,
         0.67291959,  0.36321019],
       [ 0.83302474,  0.97572153,  0.85978823, ...,  1.3599679 ,
         0.66760982,  0.36034423],
       [ 0.83962283,  0.98344987,  0.86659831, ...,  1.34923687,
         0.66234195,  0.36319839]])

    '''
    intervals = int(T/dt)

    S = np.random.random([intervals+1,paths])
    S = -1 + 2*(S > 0.5)
    S = S*np.sqrt(dt)*vol + (r - 0.5*vol*vol)*dt
    S[0] = np.ones(paths)*np.log(s0)
    S = np.exp(np.matrix.cumsum(S,axis=0))
    return(S)

def HestonSim(s0,r,T,vol,phi,kappa,xi,dt,paths):
    '''
    Simulate the motion of an underlying stock that follows a
    standard Weiner process for T/dt steps over a specified number
    of paths with a stochastic volatility

    Parameters
    ----------
    s0 : number of any type (int, float8, float64 etc.)
        Spot value of underlying assets at current time, t
    r : number of any type (int, float8, float64 etc.)
        Risk free interest rate value
    T : number of any type (int, float8, float64 etc.)
        Time till expiration for option
    vol : number of any type (int, float8, float64 etc.)
        Volatility of underlying, implied constant till
        expiration in simple model
    phi : number of any type (int, float8, float64 etc.)
        Correlation of price to volatility
    kappa : number of any type (int, float8, float64 etc.)
        Speed of volatility adjustment
    xi : number of any type (int, float8, float64 etc.)
        Volatility of the volatility
    dt : number of any type (int, float8, float64 etc.)
        Time interval used in simulation, used for number
        of stepsstock has, and motion of Weiner process
    paths : number of type 'int'
        Number of stocks simulated, higher number of paths
        leads to greater accuracy in calculated price.

    * numpy matrices will begin to break once a large enough path is chosen
    ** similarly, if too small of a dt is chosen, the numpy matrices
    will begin to break

    Returns
    -------
    [S, volMotion] : list of numpy.array's
        A (T/dt) x paths array that holds the simulated
        stock price values
        A (T/dt) x paths array that holds the simulated
        volatility motion

        both of the form:
        [[s_0 s_0 ... s_0],
        [s_11 s_12 ... s_1paths]
        ...
        [s_(T/dt)1 s_(T/dt)2 ... s_(T/dt)paths]]

    Examples
    --------
    >>> from qcfoptions.simulation import HestonSim
    >>> s0 = 1
        r = 0.015
        T = 2
        vol = 0.25
        vol = 0.25
        phi = -0.4
        kappa = 8
        dt = 0.001
        paths = 1000
    >>> HestonSim(s0,r,T,vol,dt,paths)
    [array([[ 1.        ,  1.        ,  1.        , ...,  1.        ,
          1.        ,  1.        ],
        [ 0.99220026,  0.99220026,  0.99220026, ...,  1.00768681,
          0.99188224,  0.99201948],
        [ 0.98455083,  0.98455083,  1.00019494, ...,  0.9998787 ,
          1.00018749,  0.98419354],
        ...,
        [ 0.57934074,  0.78773209,  0.77628209, ...,  0.67412693,
          0.94029086,  0.85260689],
        [ 0.58289171,  0.79231336,  0.78272252, ...,  0.67913971,
          0.93453208,  0.85863832],
        [ 0.57946344,  0.79675135,  0.77618405, ...,  0.68403361,
          0.92858262,  0.85250678]]),
 array([[ 0.0625    ,  0.0625    ,  0.0625    , ...,  0.0625    ,
          0.0625    ,  0.0625    ],
        [ 0.06107081,  0.06107081,  0.06107081, ...,  0.05885721,
          0.06614279,  0.06392919],
        [ 0.05966948,  0.05966948,  0.06468314, ...,  0.06027327,
          0.06986109,  0.06247232],
        ...,
        [ 0.03445569,  0.03621109,  0.07259113, ...,  0.05637955,
          0.03862534,  0.04846038],
        [ 0.03738478,  0.03364863,  0.06858453, ...,  0.0550711 ,
          0.0376928 ,  0.04983118],
        [ 0.03476835,  0.03120657,  0.070033  , ...,  0.05171108,
          0.0407202 ,  0.05120868]])]

    '''
    intervals = int(T/dt)

    S = np.sqrt(dt)*(-1 + 2*(np.random.random([intervals+1,paths]) > 0.5))
    V = np.sqrt(dt)*(-1 + 2*(np.random.random([intervals+1,paths]) > 0.5))
    V = phi*S + np.sqrt(1 - phi*phi)*V

    volMotion = np.zeros([intervals+1,paths])
    volMotion[0] = vol*vol*np.ones(paths)

    for t in range(intervals):
        vt = volMotion[t]
        dvt = kappa*(vol*vol - vt)*dt + xi*np.sqrt(vt)*V[t]
        volMotion[t+1] = vt + dvt

    S = (r - 0.5*volMotion)*dt + np.sqrt(volMotion)*S
    S[0] = np.ones(paths)*np.log(s0)
    S = np.exp(np.matrix.cumsum(S,axis=0))
    return([S, volMotion])


class Simple:
    '''
    Simulate the motion of an underlying stock that follows a standard
    Weiner process for T/dt steps over a specified number of paths.

    Determine the pricing of various options using the simulated stochastic
    motion of the underlying asset.

    '''
    def __init__(self,s0,r,T,vol,dt=0.001,paths=10000):
        '''
        Parameters
        ----------
        s0 : number of any type (int, float8, float64 etc.)
            Spot value of underlying assets at current time, t
        r : number of any type (int, float8, float64 etc.)
            Risk free interest rate value
        T : number of any type (int, float8, float64 etc.)
            Time till expiration for option
        vol : number of any type (int, float8, float64 etc.)
            Volatility of underlying, implied constant till
            expiration in simple model
        dt : number of any type (int, float8, float64 etc.)
            Time interval used in simulation, used for number
            of stepsstock has, and motion of Weiner process
            Standard is 0.001, unless changed by user
        paths : number of type 'int'
            Number of stocks simulated, higher number of paths
            leads to greater accuracy in calculated price.
            Standard is 10000, unless changed by user

        * numpy matrices will begin to break once a large enough path is chosen
        ** similarly, if too small of a dt is chosen, the numpy matrices
        will begin to break

        Self
        ----
        timeMatrix : numpy.array
            A (T/dt) x paths array that holds all the time values, increasing
            from zero to time T
        S : numpy.array
            A (T/dt) x paths array that holds the simulated
            stock price values
        simtime : the time to complete the simulation, useful
            in testing efficiency for variable paths and dt

        '''
        self.s0 = s0
        self.r = r
        self.T = T
        self.vol = vol
        self.dt = dt
        self.paths = paths

        timeInt = np.matrix(np.arange(0,T+dt,dt)).transpose()
        self.timeMatrix = np.matmul(timeInt,np.matrix(np.ones(paths)))

        start = time.time()
        self.S = SimpleSim(s0,r,T,vol,dt,paths)
        self.simtime = time.time() - start

    def Euro(self,k):
        out = EuroSim(self.S,k,self.r,self.T)
        return(a)

    def AsianGeoFix(self,k):
        out = AsianGeoFixSim(self.S,k,self.r,self.T)
        return(out)

    def AsianArithFix(self,k):
        out = AsianArithFixSim(self.S,k,self.r,self.T)
        return(out)

    def AsianGeoFloat(self,m):
        out = AsianGeoFloatSim(self.S,m,self.r,self.T)
        return(out)

    def AsianArithFloat(self,m):
        out = AsianArithFloatSim(self.S,m,self.r,self.T)
        return(out)

    def Power(self,k,n):
        out = PowerSim(self.S,k,self.r,self.T,n)
        return(out)

    def PowerStrike(self,k,n):
        out = PowerStrikeSim(self.S,k,self.r,self.T,n)
        return(out)

    def AvgBarrier(self,Z):
        out = AvgBarrierSim(self.S,Z,self.r,self.timeMatrix)
        return(out)

    def NoTouchSingle(self,Z,payoutScale):
        out = NoTouchSingleSim(self.S,Z,self.r,self.T,payoutScale)
        return(out)

    def NoTouchDouble(self,Z1,Z2,payoutScale):
        out = NoTouchDoubleSim(self.S,Z1,Z1,self.r,self.T,payoutScale)
        return(out)

    def CashOrNothing(self,Z,payout):
        out = CashOrNothingSim(self.S,Z,self.r,self.T,payout)
        return(out)

class Heston:
    '''
    Simulate the motion of an underlying stock that follows a standard
    Weiner process for T/dt steps over a specified number of paths,
    with stochastic volatility.

    Determine the pricing of various options using the simulated stochastic
    motion of the underlying asset.

    '''
    def __init__(self,s0,r,T,vol,phi,kappa,xi,dt=0.001,paths=10000):
        '''
        Parameters
        ----------
        s0 : number of any type (int, float8, float64 etc.)
            Spot value of underlying assets at current time, t
        r : number of any type (int, float8, float64 etc.)
            Risk free interest rate value
        T : number of any type (int, float8, float64 etc.)
            Time till expiration for option
        vol : number of any type (int, float8, float64 etc.)
            Volatility of underlying, implied constant till
            expiration in simple model
        phi : number of any type (int, float8, float64 etc.)
            Correlation of price to volatility
        kappa : number of any type (int, float8, float64 etc.)
            Speed of volatility adjustment
        xi : number of any type (int, float8, float64 etc.)
            Volatility of the volatility
        dt : number of any type (int, float8, float64 etc.)
            Time interval used in simulation, used for number
            of stepsstock has, and motion of Weiner process
            Standard is 0.001, unless changed by user
        paths : number of type 'int'
            Number of stocks simulated, higher number of paths
            leads to greater accuracy in calculated price.
            Standard is 10000, unless changed by user

        * numpy matrices will begin to break once a large enough path is chosen
        ** similarly, if too small of a dt is chosen, the numpy matrices
        will begin to break

        Self
        ----
        timeMatrix : numpy.array
            A (T/dt) x paths array that holds all the time values, increasing
            from zero to time T
        S : numpy.array
            A (T/dt) x paths array that holds the simulated
            stock price values
        volMotion : numpy.array
            A (T/dt) x paths array that holds the simulated
            volatility motion
        simtime : the time to complete the simulation, useful
            in testing efficiency for variable paths and dt

        '''
        self.s0 = s0
        self.r = r
        self.T = T
        self.vol = vol
        self.dt = dt
        self.paths = paths
        self.phi = phi # correlation of price to volatility
        self.kappa = kappa # speed of adjustment
        self.xi = xi # volatility of volatility

        timeInt = np.matrix(np.arange(0,T+dt,dt)).transpose()
        self.timeMatrix = np.matmul(timeInt,np.matrix(np.ones(paths)))

        start = time.time()
        [self.S, self.volMotion] = HestonSim(s0,r,vol,phi,kappa,xi,dt,intervals,paths)
        self.simtime = time.time() - start

    def Euro(self,k):
        out = EuroSim(self.S,k,self.r,self.T)
        return(a)

    def AsianGeoFix(self,k):
        out = AsianGeoFixSim(self.S,k,self.r,self.T)
        return(out)

    def AsianArithFix(self,k):
        out = AsianArithFixSim(self.S,k,self.r,self.T)
        return(out)

    def AsianGeoFloat(self,m):
        out = AsianGeoFloatSim(self.S,m,self.r,self.T)
        return(out)

    def AsianArithFloat(self,m):
        out = AsianArithFloatSim(self.S,m,self.r,self.T)
        return(out)

    def Power(self,k,n):
        out = PowerSim(self.S,k,self.r,self.T,n)
        return(out)

    def PowerStrike(self,k,n):
        out = PowerStrikeSim(self.S,k,self.r,self.T,n)
        return(out)

    def AvgBarrier(self,Z):
        out = AvgBarrierSim(self.S,Z,self.r,self.timeMatrix)
        return(out)

    def NoTouchSingle(self,Z,payoutScale):
        out = NoTouchSingleSim(self.S,Z,self.r,self.T,payoutScale)
        return(out)

    def NoTouchDouble(self,Z1,Z2,payoutScale):
        out = NoTouchDoubleSim(self.S,Z1,Z1,self.r,self.T,payoutScale)
        return(out)

    def CashOrNothing(self,Z,payout):
        out = CashOrNothingSim(self.S,Z,self.r,self.T,payout)
        return(out)

